/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                         LibWin.c                        **/
/**                                                         **/
/** This file contains some commonly used Windows routines. **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996-1999                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#ifdef WINDOWS

#include "LibWin.h"
#include <Windows.h>

/** WinNewImage() ********************************************/
/** Create a new image buffer.                              **/
/*************************************************************/
int WinNewImage(Image *Img,int Width,int Height,int Depth)
{
  struct
  {
    WORD Version;
    WORD Depth;
    PALETTEENTRY Entries[256];
  } PInfo;
  struct
  {
    BITMAPINFOHEADER Header;
    RGBQUAD Entries[256];
  } BInfo;
  int J;

  /* Clear the data structure */
  Img->W=Img->H=Img->D=0;
  Img->Data=0;
  Img->hDC=0;
  Img->hBMap=0;
  Img->hPal=0;

  /* Only 2..256 depth allowed */
  if((Depth<2)||(Depth>256)) return(0);

  /* Set bitmap parameters */
  BInfo.Header.biBitCount      = 8;
  BInfo.Header.biCompression   = BI_RGB;
  BInfo.Header.biWidth         = Width;
  BInfo.Header.biHeight        = -Height;
  BInfo.Header.biSize          = 40/*sizeof(BInfo)*/;
  BInfo.Header.biPlanes        = 1;
  BInfo.Header.biSizeImage     = 0;
  BInfo.Header.biXPelsPerMeter = 0;
  BInfo.Header.biYPelsPerMeter = 0;
  BInfo.Header.biClrUsed       = Depth;
  BInfo.Header.biClrImportant  = Depth;

  /* Set palette parameters */
  PInfo.Version = 0x300;
  PInfo.Depth   = Depth;
  for(J=10;J<246;J++)
  {
    PInfo.Entries[J].peRed   = BInfo.Entries[J].rgbRed   = 0x00;
    PInfo.Entries[J].peGreen = BInfo.Entries[J].rgbGreen = 0x00;
    PInfo.Entries[J].peBlue  = BInfo.Entries[J].rgbBlue  = 0x00;
    PInfo.Entries[J].peFlags = PC_NOCOLLAPSE;
    BInfo.Entries[J].rgbReserved = 0x00;
  }

  /* Create device context */
  if(!(Img->hDC=CreateCompatibleDC(NULL))) return(0);

  /* Create bitmap */
  Img->hBMap =
    CreateDIBSection
    (Img->hDC,(BITMAPINFO *)&BInfo,DIB_RGB_COLORS,&(Img->Data),NULL,NULL);
  if(!Img->hBMap) { DeleteDC(Img->hDC);Img->hDC=0;return(0); }

  /* Create palette */
  if(!(Img->hPal=CreatePalette((LOGPALETTE *)&PInfo)))
  {
    DeleteObject(Img->hBMap);Img->hBMap=0;
    DeleteDC(Img->hDC);Img->hDC=0;
    return(0);
  }

  /* Select bitmap and palette into device context */
  SelectObject(Img->hDC,Img->hBMap);
  SelectPalette(Img->hDC,Img->hPal,FALSE);
  RealizePalette(Img->hDC);

  /* Clear, set, return success */
  memset(Img->Data,0,Width*Height);
  Img->W=Width;Img->H=Height;Img->D=Depth;
  return(1);
}

/** WinFreeImage() *******************************************/
/** Free image buffer.                                      **/
/*************************************************************/
void WinFreeImage(Image *Img)
{
  /* Free bitmap and device context */
  if(Img->hBMap) DeleteObject(Img->hBMap);
  if(Img->hPal)  DeleteObject(Img->hPal);
  if(Img->hDC)   DeleteDC(Img->hDC);

  /* Clear the data structure */
  Img->W=Img->H=0;
  Img->Data=0;
  Img->hDC=0;
  Img->hBMap=0;
  Img->hPal=0;
}

/** WinPutImage() ********************************************/
/** Put an image on the screen.                             **/
/*************************************************************/
void WinPutImage(HWND hWnd,Image *Img,int X,int Y,int W,int H)
{
  HDC DestDC;
  RECT R;
  int DW,DH;

  /* If incorrect parameters, fall out */
  if((X+W>Img->W)||(Y+H>Img->H)) return;

  /* Get window's context */
  DestDC=GetDC(hWnd);

  /* Query window's size */
  GetClientRect(hWnd,(LPRECT)&R);
  DW=R.right-R.left;
  DH=R.bottom-R.top;

  /* Here goes the copy itself */
  StretchBlt(DestDC,0,0,DW,DH,Img->hDC,X,Y,W,H,SRCCOPY);

  /* We are done with the window */
  ReleaseDC(hWnd,DestDC);
}

/** WinSetColor() ********************************************/
/** Set given color to a given (R,G,B) value.               **/
/*************************************************************/
void WinSetColor(Image *Img,int N,int R,int G,int B)
{
  PALETTEENTRY PColor;
  RGBQUAD BColor;
  if(Img->hPal&&(N<Img->D))
  {
    PColor.peRed   = BColor.rgbRed   = R;
    PColor.peGreen = BColor.rgbGreen = G;
    PColor.peBlue  = BColor.rgbBlue  = B;
    PColor.peFlags = PC_NOCOLLAPSE;
    BColor.rgbReserved = 0;
    SetPaletteEntries(Img->hPal,N,1,&PColor);
    SetDIBColorTable(Img->hDC,N,1,&BColor);
  }
}

/** WinNewWindow() *******************************************/
/** Create a window of given size with a given title.       **/
/*************************************************************/
HWND WinNewWindow(HINSTANCE hInst,char *Title,int Width,int Height)
{
  return
  (
    CreateWindow
    (
      Title,Title,WS_VISIBLE|WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT,CW_USEDEFAULT,Width,Height,
      0,0,hInst,0
    )
  );
}

/** WinSizeWindow() ******************************************/
/** Change size of a window. Returns 0 on failure.          **/
/*************************************************************/
int WinSizeWindow(HWND hWnd,int Width,int Height)
{
  RECT R;

  R.top=R.left=0;R.right=Width-1;R.bottom=Height-1;
  if(!AdjustWindowRect(&R,WS_OVERLAPPEDWINDOW,TRUE)) return(0);
  Width=R.right-R.left+1;Height=R.bottom-R.top+1;
  return(SetWindowPos(hWnd,0,0,0,Width,Height,SWP_NOACTIVATE|SWP_NOZORDER|SWP_NOMOVE));
}

/** WinMoveWindow() ******************************************/
/** Change position of a window. Returns 0 on failure.      **/
/*************************************************************/
int WinMoveWindow(HWND hWnd,int X,int Y)
{
  return(SetWindowPos(hWnd,0,X,Y,0,0,SWP_NOACTIVATE|SWP_NOZORDER|SWP_NOSIZE));
}

/** WinRequestFile() *****************************************/
/** Request a file. Returns 0 on failure. The length of the **/
/** File buffer should be at least 256 characters.          **/
/*************************************************************/
int WinRequestFile(char *File,char *Types)
{
  OPENFILENAME FNInit;

  FNInit.lStructSize=sizeof(OPENFILENAME);
  FNInit.hwndOwner=GetActiveWindow();
  FNInit.hInstance=0;
  FNInit.lpstrFilter=Types;
  FNInit.lpstrCustomFilter=0;
  FNInit.nMaxCustFilter=0;
  FNInit.nFilterIndex=0;
  FNInit.lpstrFile=File;
  FNInit.nMaxFile=250;
  FNInit.lpstrFileTitle=0;
  FNInit.nMaxFileTitle=0;
  FNInit.lpstrInitialDir=0;
  FNInit.lpstrTitle=0;
  FNInit.Flags=OFN_FILEMUSTEXIST|OFN_PATHMUSTEXIST;
  FNInit.nFileOffset=0;
  FNInit.nFileExtension=0;
  FNInit.lpstrDefExt=0;
  FNInit.lCustData=0;
  FNInit.lpfnHook=0;
  FNInit.lpTemplateName=0;

  return(GetOpenFileName(&FNInit));
}

#endif /* WINDOWS */