/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                          SndWin.c                       **/
/**                                                         **/
/** This file contains standard sound generation routines   **/
/** for Windows using waveOut and midiOut.                  **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996-1999                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#ifdef WINDOWS

#include "SndWin.h"

#define SND_MSG_VOLUME 0x1000
#define SND_MSG_FREQ   0x1001
#define SND_MSG_SOUND  0x1002

static struct { BYTE Note;WORD Wheel; } Freqs[4096] =
{
#include "MIDIFreq.h"
};

static int MasterVolume = 200;
static int MasterSwitch = 0xFF;

static int SndRate      = 0;
static int NoiseGen;

static HWAVEOUT hWave   = 0;
static HMIDIOUT hMIDI   = 0;
static DWORD ThreadID   = 0;

static unsigned char Buf[SND_BUFFERS][SND_BUFSIZE];
static WAVEHDR Header[SND_BUFFERS];
static int Volume[SND_CHANNELS];
static int Freq[SND_CHANNELS];
static int Type[SND_CHANNELS];
static int Note[SND_CHANNELS];
static int Pitch[SND_CHANNELS];
static int Level[SND_CHANNELS];
static int Count[SND_CHANNELS];
static int Wave[SND_BUFSIZE];

static DWORD ThreadFunc(LPDWORD Param);
static int InitWave(HINSTANCE hInst,int Rate,int Buffers);
static int InitMIDI(HINSTANCE hInst);
static void NoteOn(int Channel,int NewNote);
static void NoteOff(int Channel);
static void ChangePitch(int Channel,int NewPitch);
static void ChangeLevel(int Channel,int NewLevel);

/** InitSound() **********************************************/
/** Initialize sound. Returns 0 on failure, effective rate  **/
/** otherwise. Special cases of rate argument: 0 - return 1 **/
/** and be silent, 1..8191 - use MIDI mapper to play music. **/
/*************************************************************/
int InitSound(HINSTANCE hInst,int Rate,int Buffers)
{
  int J;

  /* If initialized, shut down */
  TrashSound();

  /* Reset variables */
  MasterVolume=200;
  MasterSwitch=0xFF;
  SndRate=0;
  ThreadID=0;
  hWave=0;
  hMIDI=0;
  NoiseGen=1;

  /* No sound requested: EXIT */
  if(!Rate) return(1);

  /* Reset channels */
  for(J=0;J<SND_CHANNELS;J++)
  {
    Freq[J]=Volume[J]=Count[J]=0;
    Note[J]=Pitch[J]=Level[J]=0;
    Type[J]=SND_MELODIC;
  }

  return(Rate<8192? InitMIDI(hInst):InitWave(hInst,Rate,Buffers));
}

/** TrashSound() *********************************************/
/** Close all devices and free all allocated resources.     **/
/*************************************************************/
void TrashSound(void)
{
  /* Sound off */
  SndRate=0;

  /* Close waveOut if necessary */
  if(hWave)
    while(waveOutClose(hWave)==WAVERR_STILLPLAYING)
      waveOutReset(hWave);

  /* Close midiOut if necessary */
  if(hMIDI) { midiOutReset(hMIDI);midiOutClose(hMIDI); }

  hWave=0;hMIDI=0;
}

/** SetSound() ***********************************************/
/** Set sound type for a given channel.                     **/
/*************************************************************/
void SetSound(int Channel,int NewType)
{
  static int Programs[4] = { 80,126,125,80 };

  Type[Channel]=NewType;
  if(!SndRate) return;
  if(SndRate<8192)
    midiOutShortMsg
    (hMIDI,(Programs[NewType&0x03]<<8)|(Channel&0x0F)|0xC0);
}

/** SetFreq() ************************************************/
/** Set sound frequency for a given channel.                **/
/*************************************************************/
void SetFreq(int Channel,int NewFreq)
{
  if(!SndRate) return;
  if(SndRate<8192) Sound(Channel,NewFreq,Volume[Channel]);
  else PostThreadMessage(ThreadID,SND_MSG_FREQ,Channel,NewFreq);
}

/** SetVolume() **********************************************/
/** Set sound volume for a given channel.                   **/
/*************************************************************/
void SetVolume(int Channel,int NewVolume)
{
  if(!SndRate) return;
  if(SndRate<8192) Sound(Channel,Freq[Channel],NewVolume);
  else PostThreadMessage(ThreadID,SND_MSG_VOLUME,Channel,NewVolume);
}

/** Sound() **************************************************/
/** Set sound volume and frequency for a given channel.     **/
/*************************************************************/
void Sound(int Channel,int NewFreq,int NewVolume)
{
  if(!SndRate) return;
  NewVolume&=0xFF;

  if(SndRate>=8192)
    PostThreadMessage
    (ThreadID,SND_MSG_SOUND,Channel,NewVolume+(NewFreq<<8));
  else
  {
    /* Clip the frequency */
    if((NewFreq<3*3)||(NewFreq>=4096*3)) NewFreq=0;

    /* Store new frequency/volume */
    Freq[Channel]=NewFreq;Volume[Channel]=NewVolume;

    /* MIDI volume is 0..127, frequency is 0..4095 */
    NewVolume/=2;NewFreq/=3;

    /* Only if channel is on */
    if(MasterSwitch&(1<<Channel))
      if(!NewVolume||!NewFreq) NoteOff(Channel);
      else
      {
        /* Important: volume/pitch changes done when note off */
        if(Note[Channel]!=Freqs[NewFreq].Note) NoteOff(Channel);

        /* Set volume and pitch */
        ChangeLevel(Channel,NewVolume);
        ChangePitch(Channel,Freqs[NewFreq].Wheel);

        /* Start playing a note */
        NoteOn(Channel,Freqs[NewFreq].Note);
      }
  }
}

/** SetChannels() ********************************************/
/** Set overall sound volume and switch channels on/off.    **/
/*************************************************************/
void SetChannels(int Volume,int Switch)
{
  int J;

  if(!SndRate) return;

  Volume&=0xFF;
  MasterVolume=Volume;
  Switch&=((1<<SND_CHANNELS)-1);

  if(SndRate>=8192)
  {
    if(!Switch&&MasterSwitch) waveOutPause(hWave);
    if(Switch&&!MasterSwitch) waveOutRestart(hWave);
    waveOutSetVolume((UINT)hWave,Volume*0x1000100);
  }
  else
  {
    for(J=0;J<SND_CHANNELS;J++)
      if(MasterSwitch&(Switch^MasterSwitch)&(1<<J)) NoteOff(J);
    midiOutSetVolume((UINT)hMIDI,Volume*0x1000100);
  }

  MasterSwitch=Switch;
}

/** NoteOn() *************************************************/
/** Play a new note.                                        **/
/*************************************************************/
void NoteOn(int Channel,int NewNote)
{
  if(NewNote!=Note[Channel])
  {
    if(Note[Channel]) NoteOff(Channel);
    if(NewNote)
      midiOutShortMsg(hMIDI,((NewNote&0x7F)<<8)|(Channel&0x0F)|0x7F0090);
    Note[Channel]=NewNote;
  }
}

/** NoteOff() ************************************************/
/** Turn off a note.                                        **/
/*************************************************************/
void NoteOff(int Channel)
{
  if(Note[Channel])
  {
    midiOutShortMsg
    (hMIDI,((Note[Channel]&0x7F)<<8)|(Channel&0x0F)|0x7F0080);
    Note[Channel]=0;
  }
}

/** ChangePitch() ********************************************/
/** Change channel pitch.                                   **/
/*************************************************************/
void ChangePitch(int Channel,int NewPitch)
{
  if(NewPitch!=Pitch[Channel])
  {
    midiOutShortMsg
    (hMIDI,((NewPitch&0x3F80)<<1)|((NewPitch&0x7F)<<8)|(Channel&0x0F)|0xE0);
    Pitch[Channel]=NewPitch;
  }
}

/** ChangeLevel() ********************************************/
/** Change channel level (volume).                          **/
/*************************************************************/
void ChangeLevel(int Channel,int NewLevel)
{
  if(NewLevel!=Level[Channel])
  {
    midiOutShortMsg(hMIDI,((NewLevel&0x7F)<<16)|(Channel&0x0F)|0x07B0);
    Level[Channel]=NewLevel;
  }
}

/** InitWave() ***********************************************/
/** Initialize waveOut. Returns 0 on failure or effective   **/
/** rate otherwise.                                         **/
/*************************************************************/
int InitWave(HINSTANCE hInst,int Rate,int Buffers)
{
  PCMWAVEFORMAT Format;
  HANDLE hThread;
  int J;

  /* No sound yet */
  SndRate=0;

  /* Adjust number of buffers if it is wrong */
  if(Buffers<2) Buffers=2;
  else if(Buffers>SND_BUFFERS) Buffers=SND_BUFFERS;

  /* Reset headers */
  for(J=0;J<SND_BUFFERS;J++)
    Header[J].dwFlags=WHDR_DONE;

  /* Create a thread */
  hThread=CreateThread(0,0,(LPTHREAD_START_ROUTINE)ThreadFunc,0,0,&ThreadID);
  if(!hThread) return(0);

  /* Sound update thread is time critical */
  SetThreadPriority(hThread,THREAD_PRIORITY_TIME_CRITICAL);

  /* Let the thread go into message loop */
  Sleep(1000);

  /* Open WaveOut device */
  Format.wf.wFormatTag      = WAVE_FORMAT_PCM;
  Format.wf.nChannels       = 1;
  Format.wf.nSamplesPerSec  = Rate;
  Format.wBitsPerSample     = 8;
  Format.wf.nBlockAlign     = (Format.wf.nChannels*Format.wBitsPerSample)/8;
  Format.wf.nAvgBytesPerSec = Format.wf.nSamplesPerSec*Format.wf.nBlockAlign;
  if(waveOutOpen(&hWave,WAVE_MAPPER,&Format,ThreadID,0,CALLBACK_THREAD))
  { PostThreadMessage(ThreadID,MM_WOM_CLOSE,(WPARAM)hWave,0);return(0); }

  /* waveOut opened successfully */
  SndRate=Rate;

  /* Initially, pump in from all buffers */
  for(J=0;J<Buffers;J++)
    PostThreadMessage(ThreadID,MM_WOM_DONE,(WPARAM)hWave,(LPARAM)&Header[J]);

  /* Success! */
  return(SndRate);
}

/** InitMIDI() ***********************************************/
/** Initialize midiOut. Returns 0 on failure or 1 otherwise **/
/*************************************************************/
int InitMIDI(HINSTANCE hInst)
{
  DWORD Data;
  int J;

  /* No sound yet */
  SndRate=0;

  /* Open MIDI mapper */
  if(midiOutOpen(&hMIDI,MIDI_MAPPER,0,0,0)) return(0);

  /* midiOut opened successfully */
  SndRate=1;

  /* Reset MIDI */
  midiOutReset(hMIDI);
  for(J=0;J<SND_CHANNELS;J++)
    midiOutShortMsg(hMIDI,(80<<8)|(J&0x0F)|0xC0);

  /* Success! */
  return(SndRate);
}

/** ThreadFunc() *********************************************/
/** This is a thread responsible for receiving messages     **/
/** from waveOut and pumping new waveform data into it.     **/
/*************************************************************/
DWORD ThreadFunc(LPDWORD Param)
{
  register int N,J,K,I,L1,L2,V;
  HWAVEOUT hWave;
  MSG Msg;

  while(GetMessage(&Msg,0,0,0)>0)
    switch(Msg.message)
    {
      case SND_MSG_VOLUME:
        Volume[Msg.wParam]=Msg.lParam&0xFF;
        break;

      case SND_MSG_FREQ:
        Freq[Msg.wParam]=Msg.lParam>SndRate/2? 0:Msg.lParam;
        break;

      case SND_MSG_SOUND:
        Volume[Msg.wParam]=Msg.lParam&0xFF;
        Msg.lParam>>=8;
        Freq[Msg.wParam]=Msg.lParam>SndRate/2? 0:Msg.lParam;
        break;

      case MM_WOM_CLOSE:
        return(0);

      case MM_WOM_DONE:
        /* Sound must be properly initialized */
        if(SndRate<8192) break;

        /* Get header number and waveOut handle */
        N=(WAVEHDR *)(Msg.lParam)-&Header[0];
        hWave=(HWAVEOUT)(Msg.wParam);

        /* Waveform generator */
        for(J=0;J<SND_CHANNELS;J++)
          if(Freq[J]&&(V=Volume[J])&&(MasterSwitch&(1<<J)))
            switch(Type[J])
            {
              case SND_MELODIC: /* Melodic Sound */
                K=0x10000*Freq[J]/SndRate;
                L1=Count[J];
                for(I=0;I<SND_BUFSIZE;I++)
                {
                  L2=L1+K;
                  Wave[I]+=
                    V*(L1&0x8000? (L2&0x8000? 32767:0):(L2&0x8000? 0:-32767));
                  L1=L2;
                }
                Count[J]=L1;
                break;
              case SND_NOISE: /* White Noise */
                K=0x10000*Freq[J]/SndRate;
                L1=Count[J];
                for(I=0;I<SND_BUFSIZE;I++)
                {
                  L1+=K;
                  if(L1&0xFFFF0000)
                  {
                    L1&=0xFFFF;
                    if((NoiseGen<<=1)&0x80000000) NoiseGen^=0x08000001;
                  }
                  Wave[I]+=V*(NoiseGen&1? 32767:-32767);
                }
                Count[J]=L1;
                break;
            }

        /* Mix and convert waveforms */
        for(J=0;J<SND_BUFSIZE;J++)
        {
          I=Wave[J]/SND_CHANNELS/256;
          Buf[N][J]=I/256+128;
          Wave[J]=0;
        }

        /* Send buffer to the wave device */
        waveOutUnprepareHeader(hWave,&Header[N],sizeof(WAVEHDR));
        Header[N].lpData         = Buf[N];
        Header[N].dwBufferLength = SND_BUFSIZE;
        Header[N].dwFlags        = 0;
        Header[N].dwLoops        = 0;
        if(!waveOutPrepareHeader(hWave,&Header[N],sizeof(WAVEHDR)))
          waveOutWrite(hWave,&Header[N],sizeof(WAVEHDR));
    }

  return(0);
}

#endif /* WINDOWS */